@@ -18,6 +18,7 @@ import datetime
 from datetime import time
 import importlib.resources
 import json
+from typing import Optional
 
 import daiquiri
 from rfc3986 import uri_reference, validators
@@ -75,6 +76,15 @@ class Rule(object):
     processing content validation.
     """
 
+    def __init__(self, rule_name):
+        self._name = rule_name
+        rule_data = rules_dict[rule_name]
+        self._attributes = rule_data[0]
+        self._children = rule_data[1]
+        self._content = rule_data[2]
+        self._rule_children_names = self._get_rule_children_names(self._children)
+        self._depth = 0
+
     @staticmethod
     def child_list_node_names(child_list: list):
         if list is None or len(child_list) < 3:
@@ -188,18 +198,12 @@ class Rule(object):
                 logger.debug(ex)
         return is_valid
 
-    def __init__(self, rule_name):
-        self._name = rule_name
-        rule_data = rules_dict[rule_name]
-        self._attributes = rule_data[0]
-        self._children = rule_data[1]
-        self._content = rule_data[2]
-        self._named_children = self._get_children(self._children)
-
     def child_insert_index(self, parent: Node, new_child: Node) -> int:
         """
         Determines the index location of a new child node in a given parent node
-        based on the parent node's rule type
+        based on the parent node's rule type. Note that only a legal position in
+        the list of possible children is guaranteed; parent node validity may be
+        based on other constraint of the rule type.
 
         Args:
             parent: Parent node of which to be adding child node
@@ -209,12 +213,34 @@ class Rule(object):
             int: Index location of new child node
 
         """
-        new_child_position = self._get_child_position(new_child)
-        for index in range(len(parent.children) - 1, -1, -1):
-            child_position = self._get_child_position(parent.children[index])
-            if new_child_position >= child_position:
-                return index + 1
-        return 0
+        try:
+            new_child_index = self._rule_children_names.index(new_child.name)
+        except ValueError as e:
+            msg = f"Child '{new_child.name}' not allowed in parent '{parent.name}'"
+            raise ChildNotAllowedError(msg)
+        for index, child in enumerate(parent.children):
+            parent_child_index = self._rule_children_names.index(child.name)
+            if parent_child_index > new_child_index:
+                return index
+        index = len(parent.children)
+        return index
+
+    @staticmethod
+    def _is_in_path(rule_children: list, node: Node) -> bool:
+        is_in_path = False
+        if Rule._is_rule_child(rule_children):
+            return rule_children[0] == node.name
+        elif Rule._is_sequence(rule_children):
+            for child in rule_children:
+                is_in_path = Rule._is_in_path(child, node)
+                if is_in_path:
+                    break
+        else:
+            for child in rule_children[:-2]:
+                is_in_path = Rule._is_in_path(child, node)
+                if is_in_path:
+                    break
+        return is_in_path
 
     def is_required_attribute(self, attribute: str):
         if attribute in self._attributes:
@@ -223,7 +249,7 @@ class Rule(object):
             raise Exception(f"Unknown attribute {attribute}")
 
     def is_allowed_child(self, child_name: str):
-        return child_name in self._named_children
+        return child_name in self._rule_children_names
 
     def allowed_attribute_values(self, attribute: str):
         values = []
@@ -243,7 +269,7 @@ class Rule(object):
         content, attributes, and children.
 
         Args:
-            node: Node instance to be validates
+            node: Node instance to be validated
             errs: List of validation errors
 
         Returns:
@@ -252,33 +278,16 @@ class Rule(object):
         Raises:
             MetapypeRuleError: Illegal attribute or missing required attribute
         """
-        self._validate_content(node, errs)
+        if self.name in (RULE_TEXT, RULE_ANYNAME):
+            is_mixed_content = True
+        else:
+            is_mixed_content = False
+
+        self._validate_content(node, is_mixed_content, errs)
         self._validate_attributes(node, errs)
-        self._validate_children(node, errs)
+        self._validate_children(node, is_mixed_content, errs)
 
-    def _get_child_position(self, node: Node):
-        """
-        Returns the permitted position of a child node for a given rule type
-
-        Args:
-            node: Child node seeking position in sequence of allowable children
-
-        Returns:
-            position: Integer value of child node position
-
-        Raises:
-            ChildNotAllowedError: Child node not permitted by given rule type
-        """
-        for position, child in enumerate(self.children):
-            child_name_list = child[:-2]
-            if isinstance(child_name_list[0], list):
-                child_name_list = [_[0] for _ in child_name_list]
-            if node.name in child_name_list:
-                return position
-        msg = f'Child "{node.name}" not allowed'
-        raise ChildNotAllowedError(msg)
-
-    def _validate_content(self, node: Node, errs: list = None):
+    def _validate_content(self, node: Node, is_mixed_content: bool, errs: list = None):
         """
         Validates node content for rule compliance.
         For each of the content rules configured for this rule,
@@ -306,7 +315,7 @@ class Rule(object):
             elif content_rule == "intContent":
                 self._validate_int_content(node, errs)
             elif content_rule == "nonEmptyContent":
-                self._validate_non_empty_content(node, errs)
+                self._validate_non_empty_content(node, is_mixed_content, errs)
             elif content_rule == "strContent":
                 self._validate_str_content(node, errs)
             elif content_rule == "timeContent":
@@ -431,9 +440,8 @@ class Rule(object):
         self._validate_float_range_content(node, (-90.0, 90.0), errs)
 
     @staticmethod
-    def _validate_non_empty_content(node: Node, errs: list = None):
-        # if len(node.children) == 0 and node.content is None:
-        if node.content is None or len(str(node.content)) == 0:
+    def _validate_non_empty_content(node: Node, is_mixed_content: bool, errs: list = None):
+        if (node.content is None or len(str(node.content)) == 0) and not is_mixed_content:
             msg = f'Node "{node.name}" content should not be empty'
             if errs is None:
                 raise MetapypeRuleError(msg)
@@ -593,7 +601,7 @@ class Rule(object):
                             )
                         )
 
-    def _validate_children(self, node: Node, errs: list = None) -> None:
+    def _validate_children(self, node: Node, is_mixed_content: bool, errs: list = None) -> None:
         """
         Validates node children for rule compliance.
 
@@ -602,9 +610,6 @@ class Rule(object):
         3. Iterates through the list children rules and validates whether
            the node instance complies with the rules.
 
-        Args:
-            node: Node instance to be validated
-
         Returns:
             None
 
@@ -612,166 +617,104 @@ class Rule(object):
             MetapypeRuleError: Illegal child, bad sequence or choice, missing
             child, or wrong child cardinality
         """
-        if node.name == names.METADATA:
+        self._node = node
+        self._node_children_names = []
+        self._node_index = 0
+
+        if self._node.name == names.METADATA:
             # Metadata nodes may contain any type of child node, but only one such node
-            if len(node.children) > 1:
+            if len(self._node.children) > 1:
                 msg = f"Maximum occurrence of 1 child exceeded in parent '{names.METADATA}'"
                 if errs is None:
                     raise MaxOccurrenceExceededError(msg)
                 else:
-                    errs.append((ValidationError.MAX_OCCURRENCE_EXCEEDED, msg, node))
+                    errs.append((ValidationError.MAX_OCCURRENCE_EXCEEDED, msg, self._node))
+
         else:
-            node_children = list()
-            for node_child in node.children:
-                node_children.append(node_child.name)
-            # Test for illegal children nodes
-            for name in node_children:
-                if not self.is_allowed_child(name):
-                    msg = f"Child '{name}' not allowed in parent '{node.name}'"
+            for node_child in self._node.children:
+                self._node_children_names.append(node_child.name)
+
+            # Test for non-valid children
+            for node_child_name in self._node_children_names:
+                if not self.is_allowed_child(node_child_name):
+                    msg = f"Child '{node_child_name}' not allowed in parent '{self._node.name}'"
                     if errs is None:
                         raise ChildNotAllowedError(msg)
                     else:
-                        errs.append((ValidationError.CHILD_NOT_ALLOWED, msg, node, name))
+                        errs.append((ValidationError.CHILD_NOT_ALLOWED, msg, self._node, node_child_name))
 
-            if self._name in (RULE_TEXT, RULE_ANYNAME):
-                is_mixed_content = True
-            else:
-                is_mixed_content = False
+            if len(self._children) > 0:
+                # Begin validation of children
+                modality = Rule._get_children_modality(self._children)
+                if modality == "sequence":
+                    self._validate_sequence(self._children, is_mixed_content, errs)
+                else:
+                    self._validate_choice(self._children, is_mixed_content, errs)
 
-            modality = self._get_child_modality(self._children)
-            if modality == "sequence":
-                Rule._validate_children_sequence(node, node_children, self._children, is_mixed_content, errs)
-            elif modality == "choice":
-                Rule._validate_children_choice(node, node_children, self._children, is_mixed_content, errs)
-
-    @staticmethod
-    def _validate_children_sequence(
-        node: Node, node_children: list, rule_children: list, is_mixed_content: bool, errs: list = None
-    ):
-        index = 0
-        for rule_child in rule_children:
-            if Rule._get_child_modality(rule_child) == "choice":
-                index += Rule._validate_children_choice(
-                    node, node_children[index:], rule_child, is_mixed_content, errs
+            if self._node_index != len(self._node_children_names):
+                msg = (
+                    f"Child '{self._node_children_names[self._node_index]}' "
+                    f"is not allowed in this position for parent '{self._node.name}'"
                 )
-            else:
-                name = rule_child[0]
-                min = rule_child[-2]
-                max = rule_child[-1]
-                occurrence = 0
-                while (
-                    index < len(node_children) and name == node_children[index]
-                ):
-                    occurrence += 1
-                    if max is not INFINITY and occurrence > max:
-                        msg = (
-                            f"Maximum occurrence of '{max}' exceeded for child '{node_children[index]}' in parent "
-                            f"'{node.name}'"
-                        )
-                        if errs is None:
-                            raise MaxOccurrenceExceededError(msg)
-                        else:
-                            errs.append(
-                                (
-                                    ValidationError.MAX_OCCURRENCE_EXCEEDED,
-                                    msg,
-                                    node,
-                                    node_children[index],
-                                    max,
-                                )
-                            )
-                    index += 1
-                if occurrence < min:
-                    msg = f"Minimum occurrence of '{min}' not met for child '{name}' in parent '{node.name}'"
-                    if errs is None:
-                        raise MinOccurrenceUnmetError(msg)
-                    else:
-                        errs.append(
-                            (
-                                ValidationError.MIN_OCCURRENCE_UNMET,
-                                msg,
-                                node,
-                                node_children[index],
-                                min,
-                            )
-                        )
-        if index != len(node_children):
-            msg = f"Child '{node_children[index]}' is not allowed in this position for parent '{node.name}'"
-            if errs is None:
-                raise ChildNotAllowedError(msg)
-            else:
-                errs.append(
-                    (
-                        ValidationError.CHILD_NOT_ALLOWED,
-                        msg,
-                        node,
-                        node_children[index],
-                    )
-                )
-
-    @staticmethod
-    def _validate_children_choice(
-        node: Node, node_children: list, rule_children: list, is_mixed_content: bool, errs: list = None
-    ) -> int:
-        index = 0
-        choice_min = rule_children[-2]
-        choice_max = rule_children[-1]
-        choice_occurrence = 0
-        while True:
-            if choice_max is not INFINITY and choice_occurrence > choice_max:
-                msg = f"Maximum occurrence of '{choice_max}' exceeded for choice in parent '{node.name}'"
                 if errs is None:
-                    raise MaxOccurrenceExceededError(msg)
+                    raise ChildNotAllowedError(msg)
                 else:
                     errs.append(
                         (
-                            ValidationError.MAX_OCCURRENCE_EXCEEDED,
+                            ValidationError.CHILD_NOT_ALLOWED,
                             msg,
-                            node,
-                            "Choice",
-                            choice_max,
+                            self._node,
+                            self._node_children_names[self._node_index],
                         )
                     )
+
+    def _validate_sequence(self, rule_children: list, is_mixed_content: bool, errs: list = None):
+        for rule_child in rule_children:
+            if Rule._get_children_modality(rule_child) == "choice":
+                self._validate_choice(rule_child, is_mixed_content, errs)
+            else:
+                self._validate_rule_child(rule_child, False, errs)
+
+    def _validate_choice(self, rule_children: list, is_mixed_content: bool, errs: list = None):
+        choice_min = rule_children[-2]
+        choice_max = rule_children[-1]
+        choice_occurrence = 0
+        while (
+                self._node_index < len(self._node_children_names) and
+                self._node_children_names[self._node_index] in self._get_rule_children_names(rule_children)
+        ):
             for rule_child in rule_children[:-2]:
-                if Rule._get_child_modality(rule_child) == "sequence":
-                    rule_children_names = Rule._get_children(rule_child)
-                    if index < len(node_children) and node_children[index] in rule_children_names:
-                        Rule._validate_children_sequence(node, node_children, rule_child, is_mixed_content, errs)
-                        choice = True
+                if self._node_index == len(self._node_children_names):
+                    break
+                modality = Rule._get_children_modality(rule_child)
+                if modality == "sequence":
+                    if self._node_children_names[self._node_index] in self._get_rule_children_names(rule_child):
+                        self._validate_sequence(rule_child, is_mixed_content, errs)
+                        choice_occurrence += 1
+                elif modality == "choice":
+                    if self._node_children_names[self._node_index] in self._get_rule_children_names(rule_child):
+                        self._validate_choice(rule_child, is_mixed_content, errs)
                         choice_occurrence += 1
                 else:
-                    name = rule_child[0]
-                    min = rule_child[-2]
-                    max = rule_child[-1]
-                    occurrence = 0
-                    choice = False
-                    while index < len(node_children) and name == node_children[index]:
-                        index += 1
-                        occurrence += 1
-                        choice = True
-                        if max is not INFINITY and occurrence == max:
-                            break
-                    if choice:
-                        if occurrence < min:
-                            msg = f"Minimum occurrence of '{min}' not met for child '{name}' in parent '{node.name}'"
-                            if errs is None:
-                                raise MinOccurrenceUnmetError(msg)
-                            else:
-                                errs.append(
-                                    (
-                                        ValidationError.MIN_OCCURRENCE_UNMET,
-                                        msg,
-                                        node,
-                                        name,
-                                        min,
-                                    )
-                                )
+                    if self._node_children_names[self._node_index] == rule_child[0]:
+                        self._validate_rule_child(rule_child, True, errs)
                         choice_occurrence += 1
-                        break
-            if not choice:
-                break
-        if not choice and choice_occurrence < choice_min and not is_mixed_content:
-            msg = f"Minimum occurrence of '{choice_min}' not met for choice in parent '{node.name}'"
+        if choice_max is not INFINITY and choice_occurrence > choice_max:
+            msg = f"Maximum occurrence of '{choice_max}' exceeded for choice in parent '{self._node.name}'"
+            if errs is None:
+                raise MaxOccurrenceExceededError(msg)
+            else:
+                errs.append(
+                    (
+                        ValidationError.MAX_OCCURRENCE_EXCEEDED,
+                        msg,
+                        self._node,
+                        "Choice",
+                        choice_min,
+                    )
+                )
+        if choice_occurrence < choice_min and not is_mixed_content:
+            msg = f"Minimum occurrence of '{choice_min}' not met for choice in parent '{self._node.name}'"
             if errs is None:
                 raise MinOccurrenceUnmetError(msg)
             else:
@@ -779,34 +722,113 @@ class Rule(object):
                     (
                         ValidationError.MIN_OCCURRENCE_UNMET,
                         msg,
-                        node,
+                        self._node,
                         "Choice",
                         choice_min,
                     )
                 )
-        return index
+
+    def _validate_rule_child(self, rule_child: list, limit_max: bool, errs: list = None):
+        rule_child_name = rule_child[0]
+        rule_child_min = rule_child[-2]
+        rule_child_max = rule_child[-1]
+        occurrence = 0
+        while (
+                self._node_index < len(self._node_children_names) and
+                rule_child_name == self._node_children_names[self._node_index]
+        ):
+            occurrence += 1
+            self._node_index += 1
+            if limit_max and occurrence == rule_child_max:
+                return None
+            if rule_child_max is not INFINITY and occurrence > rule_child_max:
+                msg = (
+                    f"Maximum occurrence of '{rule_child_max}' "
+                    f"exceeded for child '{self._node_children_names[self._node_index]}' in parent "
+                    f"'{self._node.name}'"
+                )
+                if errs is None:
+                    raise MaxOccurrenceExceededError(msg)
+                else:
+                    errs.append(
+                        (
+                            ValidationError.MAX_OCCURRENCE_EXCEEDED,
+                            msg,
+                            self._node,
+                            None if self._node_index >= len(self._node_children_names)
+                            else self._node_children_names[self._node_index],
+                            rule_child_max,
+                        )
+                    )
+        if occurrence < rule_child_min:
+            msg = (
+                f"Minimum occurrence of '{rule_child_min}' "
+                f"not met for child '{rule_child_name}' in parent '{self._node.name}'"
+            )
+            if errs is None:
+                raise MinOccurrenceUnmetError(msg)
+            else:
+                errs.append(
+                    (
+                        ValidationError.MIN_OCCURRENCE_UNMET,
+                        msg,
+                        self._node,
+                        None if self._node_index >= len(self._node_children_names)
+                        else self._node_children_names[self._node_index],
+                        rule_child_min,
+                    )
+                )
 
     @staticmethod
-    def _get_child_modality(child: list) -> str:
-        if len(child) > 3 and not isinstance(child[-1], list):
-            mode = "choice"
-        elif len(child) == 3 and isinstance(child[0], str):
-            mode = "child"
-        else:
+    def _get_children_modality(rule_children: list) -> str:
+        if Rule._is_rule_child(rule_children):
+            mode = "child_rule"
+        elif Rule._is_sequence(rule_children):
             mode = "sequence"
+        elif Rule._is_choice(rule_children):
+            mode = "choice"
+        else:
+            msg = f"Unknown modality for {rule_children}"
+            raise ValueError(msg)
         return mode
 
     @staticmethod
-    def _get_children(children: list) -> list:
-        allowed = list()
-        if Rule._get_child_modality(children) == "choice":
-            allowed += Rule._get_children(children[:-2])
-        elif Rule._get_child_modality(children) == "sequence":
-            for child in children:
-                allowed += Rule._get_children(child)
-        else:
-            allowed.append(children[0])
-        return allowed
+    def _is_rule_child(rule_children: list) -> bool:
+        # of the form ["str", int, int|None]
+        is_rule_child = False
+        if isinstance(rule_children[0], str):
+            is_rule_child = True
+        return is_rule_child
+
+    @staticmethod
+    def _is_sequence(rule_children: list) -> bool:
+        # of the form [["str", int, int|None]...[]]]
+        is_sequence = False
+        if len(rule_children) >= 1 and isinstance(rule_children[-1], list):
+            is_sequence = True
+        return is_sequence
+
+    @staticmethod
+    def _is_choice(rule_children: list) -> bool:
+        # of the form [["str", int, int|None]...[], int, int|None]]
+        is_choice = False
+        if len(rule_children) >= 3 and (isinstance(rule_children[0], list) and isinstance(rule_children[-2], int)):
+            is_choice = True
+        return is_choice
+
+    @staticmethod
+    def _get_rule_children_names(children: list) -> list:
+        children_names = []
+        if len(children) > 0:
+            modality = Rule._get_children_modality(children)
+            if modality == "choice":
+                children_names += Rule._get_rule_children_names(children[:-2])
+            elif modality == "sequence":
+                for child in children:
+                    children_names += Rule._get_rule_children_names(child)
+            else:
+                children_names.append(children[0])
+        return children_names
 
     @property
     def name(self):
@@ -909,6 +931,7 @@ RULE_REFERENCES = "referencesRule"
 RULE_RELATEDPROJECT = "relatedProjectRule"
 RULE_RESPONSIBLEPARTY = "responsiblePartyRule"
 RULE_RESPONSIBLEPARTY_WITH_ROLE = "responsiblePartyWithRoleRule"
+RULE_ROLE = "roleRule"
 RULE_ROWCOLUMN = "rowColumnRule"
 RULE_SAMPLING = "samplingRule"
 RULE_SECTION = "sectionRule"
@@ -1112,7 +1135,7 @@ node_mappings = {
     names.RECORDDELIMITER: RULE_ANYSTRING,
     names.REFERENCES: RULE_REFERENCES,
     names.RELATED_PROJECT: RULE_RELATEDPROJECT,
-    names.ROLE: RULE_ANYSTRING,
+    names.ROLE: RULE_ROLE,
     names.ROWCOLUMNORIENTATION: RULE_ROWCOLUMN,
     names.SALUTATION: RULE_ANYNAME,
     names.SAMPLING: RULE_SAMPLING,
